# -*- coding: UTF-8 -*-

from utils import *
import selectors
import socket
from queue import Queue
import syslog

class MemoryNode:
	def __init__(self, username, sock):
		self.username = username
		self.sock = sock
		self.msg_queue = Queue()
		self.current_msg = None

class Memory:
	
	def __init__(self):
		self.by_username = {}
		self.by_socket = {}
	
	def put_in_users_msgs(self, username, element):
		if username in self.by_username:
			node = self.by_username[username]
			node.msg_queue.put(element)
			return True
		else:
			return False
	
	def create_error_message(self, addresser, addressee):
		return {'type': MessageType.INFO, 'to': addresser, 'info': f'{addressee} did not receive your message.'}
	
	def process(self, message, sock):
		if 'type' in message:
			tp = message['type']
			if tp == MessageType.GREETINGS:
				username = message['name']
				node = MemoryNode(username, sock)
				self.by_username[username] = node
				self.by_socket[sock] = node
			elif tp == MessageType.INFO:
				addressee = message['to']
				self.put_in_users_msgs(addressee, message)
			elif tp == MessageType.MESSAGE:
				addresser = message['from']
				addressee = message['to']
				msg_put = self.put_in_users_msgs(addressee, message)
				if not msg_put:
					err_msg = self.create_error_message(addresser, addressee)
					self.put_in_users_msgs(addresser, err_msg)
	
	def process_closed_socket(self, sock):
		if sock in self.by_socket:
			node = self.by_socket[sock]
			if node.current_msg is not None:
				node.msg_queue.put(node.current_msg)
			while not node.msg_queue.empty():
				msg = node.msg_queue.get()
				if msg['type'] == MessageType.MESSAGE:
					addresser = msg['from']
					err_msg = self.create_error_message(addresser, msg['to'])
					self.put_in_users_msgs(addresser, err_msg)
			del self.by_username[node.username]
			del self.by_socket[node.sock]
	
	def message_delivered(self, sock):
		if sock in self.by_socket:
			node = self.by_socket[sock]
			node.current_msg = None
	
	def next_message(self, sock):
		if sock in self.by_socket:
			node = self.by_socket[sock]
			node.current_msg = None if node.msg_queue.empty() else node.msg_queue.get()
			return node.current_msg
		else:
			return None

def accept_wrapper(sock, sel):
	conn, addr = sock.accept()
	conn.setblocking(False)
	data = SocketData()
	events = selectors.EVENT_READ | selectors.EVENT_WRITE
	sel.register(conn, events, data=data)

def log_message(message):
	if ('type' in message) and (message['type'] == MessageType.MESSAGE):
		syslog.syslog(f'FROM {message["from"]} TO {message["to"]} MESSAGE {message["text"]}')

def service_connection(key, mask, mem):
	sock = key.fileobj
	data = key.data
	if mask & selectors.EVENT_READ:
		recv_data = sock.recv(1024)
		if recv_data:
			data.inb, msgs_bytes = split_bytes(data.inb, recv_data)
			for bytes in msgs_bytes:
				msg = bytes_to_dict(bytes)
				log_message(msg)
				mem.process(msg, sock)
		else:
			sel.unregister(sock)
			sock.close()
			mem.process_closed_socket(sock)
	if mask & selectors.EVENT_WRITE:
		if data.outb:
			sent = sock.send(data.outb)
			data.outb = data.outb[sent:]
			if not data.outb:
				mem.message_delivered(sock)
		else:
			message = mem.next_message(sock)
			data.outb = b'' if message is None else dict_to_bytes(message)

if __name__ == '__main__':
	host = input('Input server address: ')
	
	sel = selectors.DefaultSelector()
	
	listening_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	listening_socket.bind((host, PORT))
	listening_socket.listen()
	listening_socket.setblocking(False)
	
	sel.register(listening_socket, selectors.EVENT_READ, data=None)
	
	mem = Memory()
	
	syslog.openlog(ident = 'chat server')
	
	while True:
		events = sel.select(timeout = None)
		for key, mask in events:
			if key.data is None:
				accept_wrapper(key.fileobj, sel)
			else:
				service_connection(key, mask, mem)



# -*- coding: UTF-8 -*-

import json as _json
from enum import Enum as _Enum, auto as _auto

PORT = 60606

class MessageType(_Enum):
	GREETINGS = _auto()
	MESSAGE = _auto()
	EXIT = _auto()
	INFO = _auto()

class SocketData:
	def __init__(self):
		self.inb = b''
		self.outb = b''

class _MessageTypeEncoder(_json.JSONEncoder):
	def default(self, obj):
		if type(obj) == MessageType:
			return {'__enum__': str(obj)}
		else:
			super.default(self, obj)

def _as_MessageType(d):
	if '__enum__' in d:
		_, member = d['__enum__'].split('.')
		return getattr(MessageType, member)
	else:
		return d

def dict_to_bytes(dic):
	json = _json.dumps(dic, cls = _MessageTypeEncoder)
	return (json + '\0').encode('utf-8')

def bytes_to_dict(bytes):
	try:
		msg = bytes.decode('utf-8')
		dump = msg[0:-1]
		return _json.loads(dump, object_hook = _as_MessageType)
	except:
		return {}

def split_bytes(buff, recved):
	result = []
	index = recved.find(b'\x00')
	while index >= 0:
		message = buff + recved[:index + 1]
		result.append(message)
		buff = b''
		recved = recved[index + 1:]
		index = recved.find(b'\x00')
	buff += recved
	return (buff, result) 

if __name__ == '__main__':
	d1 = {'name': 'Danila', 'mood': 'depressed', 'age': 20, 'msgtp': MessageType.INFO}
	
	bytes = dict_to_bytes(d1)
	
	d2 = bytes_to_dict(bytes)
	
	print(d1 == d2, d1, d2, sep = '\n')



# -*- coding: UTF-8 -*-

from utils import *
from queue import Queue, Empty
from threading import Thread
import socket, selectors

def process(message, msg_queue):
	msg = bytes_to_dict(message)
	if 'type' in msg:
		if msg['type'] == MessageType.MESSAGE:
			print(msg['from'], msg['text'], sep = ': ')
			answer = {'type': MessageType.INFO, 'to': msg['from'], 'info': f'{msg["to"]} received your message.'}
			msg_queue.put(answer)
		elif msg['type'] == MessageType.INFO:
			print(msg['info'])
	else:
		print(msg)

def process_msgs(msg_queue, sel):
	running = True
	while running:
		for key, mask in sel.select(timeout = None):
			sock = key.fileobj
			context = key.data
			if mask & selectors.EVENT_READ:
				recv_data = sock.recv(1024)
				context.inb, msgs = split_bytes(context.inb, recv_data)
				for msg in msgs:
					process(msg, msg_queue)
			if mask & selectors.EVENT_WRITE:
				if context.outb == b'':
					try:
						msg = msg_queue.get_nowait()
						if msg['type'] == MessageType.EXIT:
							running = False
							sel.unregister(sock)
							sock.close()
						else:
							context.outb = dict_to_bytes(msg)
						msg_queue.task_done()
					except Empty:
						pass
				else:
					sent = sock.send(context.outb)
					context.outb = context.outb[sent:]
	sel.close()

def start_receiving():
	msg_queue = Queue()
	
	sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	sock.setblocking(False)
	sock.connect_ex((address, PORT))
	
	sel = selectors.DefaultSelector()
	events = selectors.EVENT_READ | selectors.EVENT_WRITE
	data = SocketData()
	sel.register(sock, events, data=data)
	
	receiving_thread = Thread(target = process_msgs, args = (msg_queue, sel))
	receiving_thread.start()
	
	return msg_queue

if __name__ == '__main__':
	address = input('Input server address: ')
	name = input('Input your name: ')
	
	print('To send a message input its text and addressee.\nInput "exit" to exit.')
	
	msg_queue = start_receiving()
	
	greetings = {'type': MessageType.GREETINGS, 'name': name}
	msg_queue.put(greetings)
	
	text = input()
	while text != 'exit':
		addressee = input()
		message = {'type': MessageType.MESSAGE, 'from': name, 'to': addressee, 'text': text}
		msg_queue.put(message)
		text = input()
	
	msg_queue.put({'type': MessageType.EXIT})
	msg_queue.join()



from setuptools import setup

long_descr = ''
with open('README.md', 'r') as fh:
	long_descr = fh.read()

setup(
	name = 'console-chat-dsima',
	version = '0.0.1',
	description = 'Console chat.',
	classifiers = [
		'Programming Language :: Python :: 3.8',
		'License :: OSI Approved :: MIT License',
		'Operating System :: POSIX :: Linux'
	],
	py_modules = [
		'chat-client',
		'chat-server',
		'utils'
	],
	package_dir = {'': 'src'},
	long_description = long_descr,
	long_description_content_type = 'text/markdown',
	url = 'https://github.com/danilasimanok/ConsoleChat',
	author = 'Danila Simanok',
	author_email = 'danilasimanok@gmail.com'
)
